import scs
import numpy as np
from scipy import sparse, randn

#############################################
#      Generate random cone problems        #
#############################################


def gen_feasible(K, n, density):
  m = get_scs_cone_dims(K)
  z = randn(m,)
  y = proj_dual_cone(z, K)  # y = s - z;
  s = y - z  # s = proj_cone(z,K)

  A = sparse.rand(m, n, density, format='csc')
  A.data = randn(A.nnz)
  x = randn(n)
  c = -np.transpose(A).dot(y)
  b = A.dot(x) + s

  data = {'A': A, 'b': b, 'c': c}
  return data, np.dot(c, x)


def gen_infeasible(K, n):
  m = get_scs_cone_dims(K)

  z = randn(m,)
  y = proj_dual_cone(z, K)  # y = s - z;
  A = randn(m, n)
  A = A - np.outer(y, np.transpose(A).dot(y)) / np.linalg.norm(y)**2  # dense...

  b = randn(m)
  b = -b / np.dot(b, y)

  data = {'A': sparse.csc_matrix(A), 'b': b, 'c': randn(n)}
  return data


def gen_unbounded(K, n):
  m = get_scs_cone_dims(K)

  z = randn(m)
  s = proj_cone(z, K)
  A = randn(m, n)
  x = randn(n)
  A = A - np.outer(s + A.dot(x), x) / np.linalg.norm(x)**2
  # dense...
  c = randn(n)
  c = -c / np.dot(c, x)

  data = {'A': sparse.csc_matrix(A), 'b': randn(m), 'c': c}
  return data


def pos(x):
  return (x + abs(x)) / 2


def get_scs_cone_dims(K):
  l = K['f'] + K['l']
  for i in range(0, len(K['q'])):
    l = l + K['q'][i]

  for i in range(0, len(K['s'])):
    l = l + get_sd_cone_size(K['s'][i])

  l = l + K['ep'] * 3
  l = l + K['ed'] * 3
  l = l + len(K['p']) * 3
  return int(l)


def proj_dual_cone(z, c):
  return z + proj_cone(-z, c)


def get_sd_cone_size(n):
  return int((n * (n + 1)) / 2)


def proj_cone(z, c):
  z = np.copy(z)
  free_len = c['f']
  lp_len = c['l']
  q = c['q']
  s = c['s']
  p = c['p']
  # free/zero cone
  z[0:free_len] = 0
  # lp cone
  z[free_len:lp_len + free_len] = pos(z[free_len:lp_len + free_len])
  # SOCs
  idx = lp_len + free_len
  for i in range(0, len(q)):
    z[idx:idx + q[i]] = proj_soc(z[idx:idx + q[i]])
    idx = idx + q[i]
  # SDCs
  for i in range(0, len(s)):
    sz = get_sd_cone_size(s[i])
    z[idx:idx + sz] = proj_sdp(z[idx:idx + sz], s[i])
    idx = idx + sz
  # Exp primal
  for i in range(0, c['ep']):
    z[idx:idx + 3] = project_exp_bisection(z[idx:idx + 3])
    idx = idx + 3
  # Exp dual
  for i in range(0, c['ed']):
    z[idx:idx + 3] = z[idx:idx + 3] + project_exp_bisection(-z[idx:idx + 3])
    idx = idx + 3
  # Power
  for i in range(0, len(p)):
    if p[i] >= 0:  # primal
      z[idx:idx + 3] = proj_pow(z[idx:idx + 3], p[i])
    else:  # dual
      z[idx:idx + 3] = z[idx:idx + 3] + proj_pow(-z[idx:idx + 3], -p[i])
    idx = idx + 3
  return z


def proj_soc(tt):
  tt = np.copy(tt)
  if len(tt) == 0:
    return
  elif len(tt) == 1:
    return pos(tt)

  v1 = tt[0]
  v2 = tt[1:]
  if np.linalg.norm(v2) <= -v1:
    v2 = np.zeros(len(v2))
    v1 = 0
  elif np.linalg.norm(v2) > abs(v1):
    v2 = 0.5 * (1 + v1 / np.linalg.norm(v2)) * v2
    v1 = np.linalg.norm(v2)
  tt[0] = v1
  tt[1:] = v2
  return tt


def proj_sdp(z, n):
  z = np.copy(z)
  if n == 0:
    return
  elif n == 1:
    return pos(z)
  tidx = np.triu_indices(n)
  tidx = (tidx[1], tidx[0])
  didx = np.diag_indices(n)

  a = np.zeros((n, n))
  a[tidx] = z
  a = (a + np.transpose(a))
  a[didx] = a[didx] / np.sqrt(2.)

  w, v = np.linalg.eig(a)  # cols of v are eigenvectors
  w = pos(w)
  a = np.dot(v, np.dot(np.diag(w), np.transpose(v)))
  a[didx] = a[didx] / np.sqrt(2.)
  z = a[tidx]
  return z


def proj_pow(v, a):
  CONE_MAX_ITERS = 20
  CONE_TOL = 1e-8

  if v[0] >= 0 and v[1] >= 0 and (v[0]**a) * (v[1]**(1 - a)) >= abs(v[2]):
    return v

  if (v[0] <= 0 and v[1] <= 0 and
      ((-v[0] / a)**a) * ((-v[1] / (1 - a))**(1 - a)) >= abs(v[2])):
    return np.zeros(3,)

  xh = v[0]
  yh = v[1]
  zh = v[2]
  rh = abs(zh)
  r = rh / 2
  for iter in range(0, CONE_MAX_ITERS):
    x = calc_x(r, xh, rh, a)
    y = calc_x(r, yh, rh, 1 - a)

    f = calc_f(x, y, r, a)
    if abs(f) < CONE_TOL:
      break

    dxdr = calcdxdr(x, xh, rh, r, a)
    dydr = calcdxdr(y, yh, rh, r, (1 - a))
    fp = calc_fp(x, y, dxdr, dydr, a)

    r = min(max(r - f / fp, 0), rh)

  z = np.sign(zh) * r
  v[0] = x
  v[1] = y
  v[2] = z
  return v


def calc_x(r, xh, rh, a):
  return max(0.5 * (xh + np.sqrt(xh * xh + 4 * a * (rh - r) * r)), 1e-12)


def calcdxdr(x, xh, rh, r, a):
  return a * (rh - 2 * r) / (2 * x - xh)


def calc_f(x, y, r, a):
  return (x**a) * (y**(1 - a)) - r


def calc_fp(x, y, dxdr, dydr, a):
  return (x**a) * (y**(1 - a)) * (a * dxdr / x + (1 - a) * dydr / y) - 1


def project_exp_bisection(v):
  v = np.copy(v)
  r = v[0]
  s = v[1]
  t = v[2]
  # v in cl(Kexp)
  if (s * np.exp(r / s) <= t and s > 0) or (r <= 0 and s == 0 and t >= 0):
    return v
  # -v in Kexp^*
  if (-r < 0 and r * np.exp(s / r) <= -np.exp(1) * t) or (-r == 0 and
                                                          -s >= 0 and -t >= 0):
    return np.zeros(3,)
  # special case with analytical solution
  if r < 0 and s < 0:
    v[1] = 0
    v[2] = max(v[2], 0)
    return v

  x = np.copy(v)
  ub, lb = get_rho_ub(v)
  for iter in range(0, 100):
    rho = (ub + lb) / 2
    g, x = calc_grad(v, rho, x)
    if g > 0:
      lb = rho
    else:
      ub = rho
    if ub - lb < 1e-6:
      break
  return x


def get_rho_ub(v):
  lb = 0
  rho = 2**(-3)
  g, z = calc_grad(v, rho, v)
  while g > 0:
    lb = rho
    rho = rho * 2
    g, z = calc_grad(v, rho, z)
  ub = rho
  return ub, lb


def calc_grad(v, rho, warm_start):
  x = solve_with_rho(v, rho, warm_start[1])
  if x[1] == 0:
    g = x[0]
  else:
    g = (x[0] + x[1] * np.log(x[1] / x[2]))
  return g, x


def solve_with_rho(v, rho, w):
  x = np.zeros(3)
  x[2] = newton_exp_onz(rho, v[1], v[2], w)
  x[1] = (1 / rho) * (x[2] - v[2]) * x[2]
  x[0] = v[0] - rho
  return x


def newton_exp_onz(rho, y_hat, z_hat, w):
  t = max(max(w - z_hat, -z_hat), 1e-6)
  for iter in range(0, 100):
    f = (1 / rho**2) * t * (t + z_hat) - y_hat / rho + np.log(t / rho) + 1
    fp = (1 / rho**2) * (2 * t + z_hat) + 1 / t

    t = t - f / fp
    if t <= -z_hat:
      t = -z_hat
      break
    elif t <= 0:
      t = 0
      break
    elif abs(f) < 1e-6:
      break
  return t + z_hat



from __future__ import print_function
import platform
## import utilities to generate random cone probs:
import sys
import gen_random_cone_prob as tools


def import_error(msg):
  print()
  print('## IMPORT ERROR:' + msg)
  print()


try:
  from nose.tools import assert_raises, assert_almost_equals
except ImportError:
  import_error('Please install nose to run tests.')
  raise

try:
  import scs
except ImportError:
  import_error('You must install the scs module before running tests.')
  raise

try:
  import numpy as np
except ImportError:
  import_error('Please install numpy.')
  raise

try:
  import scipy.sparse as sp
except ImportError:
  import_error('Please install scipy.')
  raise


def check_solution(solution, expected):
  assert_almost_equals(solution, expected, places=2)


def assert_(str1, str2):
  if (str1 != str2):
    print('assert failure: %s != %s' % (str1, str2))
  assert str1 == str2


def check_infeasible(sol):
  assert_(sol['info']['status'], 'Infeasible')


def check_unbounded(sol):
  assert_(sol['info']['status'], 'Unbounded')


np.random.seed(0)
num_feas = 50
num_unb = 10
num_infeas = 10

opts = {
    'max_iters': 100000,
    'eps': 1e-5
}  # better accuracy than default to ensure test pass
K = {
    'f': 10,
    'l': 25,
    'q': [5, 10, 0, 1],
    's': [2, 1, 2, 0, 1],
    'ep': 0,
    'ed': 0,
    'p': [0.25, -0.75, 0.33, -0.33, 0.2]
}
m = tools.get_scs_cone_dims(K)


def test_feasible():
  for i in range(num_feas):
    data, p_star = tools.gen_feasible(K, n=m // 3, density=0.1)

    sol = scs.solve(data, K, use_indirect=False, **opts)
    yield check_solution, np.dot(data['c'], sol['x']), p_star
    yield check_solution, np.dot(-data['b'], sol['y']), p_star

    sol = scs.solve(data, K, use_indirect=True, **opts)
    yield check_solution, np.dot(data['c'], sol['x']), p_star
    yield check_solution, np.dot(-data['b'], sol['y']), p_star


def test_infeasible():
  for i in range(num_infeas):
    data = tools.gen_infeasible(K, n=m // 3)

    yield check_infeasible, scs.solve(data, K, use_indirect=False, **opts)
    yield check_infeasible, scs.solve(data, K, use_indirect=True, **opts)


def test_unbounded():
  for i in range(num_unb):
    data = tools.gen_unbounded(K, n=m // 2)

    yield check_unbounded, scs.solve(data, K, use_indirect=False, **opts)
    yield check_unbounded, scs.solve(data, K, use_indirect=True, **opts)



from __future__ import print_function, division
import scs
import numpy as np
from scipy import sparse, randn
import gen_random_cone_prob as tools

#############################################
#  Uses scs to solve a random cone problem  #
#############################################


def main():
  flags = [(False, False), (True, False)]
  try:
    import _scs_gpu
    flags += [(True, True)]
  except ImportError:
    pass

  for (use_indirect, gpu) in flags:
    np.random.seed(1)
    solve_feasible(use_indirect, gpu)
    solve_infeasible(use_indirect, gpu)
    solve_unbounded(use_indirect, gpu)


def solve_feasible(use_indirect, gpu):
  # cone:
  K = {
      'f': 10,
      'l': 15,
      'q': [5, 10, 0, 1],
      's': [3, 4, 0, 0, 1],
      'ep': 10,
      'ed': 10,
      'p': [-0.25, 0.5, 0.75, -0.33]
  }
  m = tools.get_scs_cone_dims(K)
  data, p_star = tools.gen_feasible(K, n=m // 3, density=0.01)
  params = {'normalize': True, 'scale': 5, 'cg_rate': 2}

  sol = scs.solve(data, K, use_indirect=use_indirect, gpu=gpu, **params)
  x = sol['x']
  y = sol['y']
  print('p*  = ', p_star)
  print('pri error = ', (np.dot(data['c'], x) - p_star) / p_star)
  print('dual error = ', (-np.dot(data['b'], y) - p_star) / p_star)


def solve_infeasible(use_indirect, gpu):
  K = {
      'f': 10,
      'l': 15,
      'q': [5, 10, 0, 1],
      's': [3, 4, 0, 0, 1],
      'ep': 10,
      'ed': 10,
      'p': [-0.25, 0.5, 0.75, -0.33]
  }
  m = tools.get_scs_cone_dims(K)
  data = tools.gen_infeasible(K, n=m // 3)
  params = {'normalize': True, 'scale': 0.5, 'cg_rate': 2}
  sol = scs.solve(data, K, use_indirect=use_indirect, gpu=gpu, **params)


def solve_unbounded(use_indirect, gpu):
  K = {
      'f': 10,
      'l': 15,
      'q': [5, 10, 0, 1],
      's': [3, 4, 0, 0, 1],
      'ep': 10,
      'ed': 10,
      'p': [-0.25, 0.5, 0.75, -0.33]
  }
  m = tools.get_scs_cone_dims(K)
  data = tools.gen_unbounded(K, n=m // 3)
  params = {'normalize': True, 'scale': 0.5, 'cg_rate': 2}
  sol = scs.solve(data, K, use_indirect=use_indirect, gpu=gpu, **params)


if __name__ == '__main__':
  main()



from __future__ import print_function
import platform
## import utilities to generate random cone probs:
import sys
import gen_random_cone_prob as tools

# if platform.system() == 'Windows':
#   print('Skipping Python linear system tests on Windows')
#   sys.exit(0)


def import_error(msg):
  print()
  print('## IMPORT ERROR:' + msg)
  print()


try:
  from nose.tools import assert_raises, assert_almost_equals
except ImportError:
  import_error('Please install nose to run tests.')
  raise

try:
  import scs
except ImportError:
  import_error('You must install the scs module before running tests.')
  raise

try:
  import numpy as np
except ImportError:
  import_error('Please install numpy.')
  raise

try:
  import scipy.sparse as sp
  import scipy.sparse.linalg as sla
except ImportError:
  import_error('Please install scipy.')
  raise


def check_solution(solution, expected):
  assert_almost_equals(solution, expected, places=2)


def assert_(str1, str2):
  if (str1 != str2):
    print('assert failure: %s != %s' % (str1, str2))
  assert str1 == str2


np.random.seed(0)
num_probs = 50

K = {
    'f': 10,
    'l': 25,
    'q': [5, 10, 0, 1],
    's': [2, 1, 2, 0, 1],
    'ep': 0,
    'ed': 0,
    'p': [0.25, -0.75, 0.33, -0.33, 0.2]
}
m = tools.get_scs_cone_dims(K)

def test_python_linsys():
  global Msolve, A, nz_cone, ncon_cone

  for i in range(num_probs):
    data, p_star = tools.gen_feasible(K, n=m // 3, density=0.1)

    A = data['A']
    ncon_cone, nz_cone = A.shape
    Msolve = None

    def init_lin_sys_work_cb(rho):
        global Msolve, nz_cone, ncon_cone, A
        M = sp.bmat([[rho*sp.eye(nz_cone), A.T], [A, -sp.eye(ncon_cone)]])
        Msolve = sla.factorized(M)

    def solve_lin_sys_cb(b, s, i):
        global Msolve
        b[:] = Msolve(b)

    def accum_by_a_cb(x, y):
        global A
        y += A.dot(x)

    def accum_by_atrans_cb(x, y):
        global A
        y += A.T.dot(x)

    def normalize_a_cb(boundaries, scale):
        global A, ncon_cone, nz_cone

        D_all = np.ones(ncon_cone)
        E_all = np.ones(nz_cone)

        min_scale, max_scale = (1e-4, 1e4)
        n_passes = 10

        for i in range(n_passes):
            D = np.sqrt(sla.norm(A, float('inf'), axis=1))
            E = np.sqrt(sla.norm(A, float('inf'), axis=0))
            D[D < min_scale] = 1.0
            E[E < min_scale] = 1.0
            D[D > max_scale] = max_scale
            E[E > max_scale] = max_scale
            start = boundaries[0]
            for delta in boundaries[1:]:
                D[start:start+delta] = D[start:start+delta].mean()
                start += delta
            A = sp.diags(1/D).dot(A).dot(sp.diags(1/E))
            D_all *= D
            E_all *= E

        mean_row_norm = sla.norm(A, 2, axis=1).mean()
        mean_col_norm = sla.norm(A, 2, axis=0).mean()
        A *= scale

        return D_all, E_all, mean_row_norm, mean_col_norm

    def un_normalize_a_cb(D, E, scale):
        global A
        A = sp.diags(D).dot(A).dot(sp.diags(E))/scale

    sol = scs.solve(
        data, K, verbose=False, use_indirect=False,
        normalize=False,
        linsys_cbs=(
            init_lin_sys_work_cb,
            solve_lin_sys_cb, accum_by_a_cb, accum_by_atrans_cb,
            normalize_a_cb, un_normalize_a_cb,
        ),
        max_iters=int(1e5), eps=1e-5,
    )

    yield check_solution, np.dot(data['c'], sol['x']), p_star
    yield check_solution, np.dot(-data['b'], sol['y']), p_star

    sol = scs.solve(
        data, K, verbose=False, use_indirect=False,
        normalize=True,
        linsys_cbs=(
            init_lin_sys_work_cb,
            solve_lin_sys_cb, accum_by_a_cb, accum_by_atrans_cb,
            normalize_a_cb, un_normalize_a_cb,
        ),
        max_iters=int(1e5), eps=1e-5,
    )

    yield check_solution, np.dot(data['c'], sol['x']), p_star
    yield check_solution, np.dot(-data['b'], sol['y']), p_star

# for i,c in zip(range(4), test_python_linsys()):
#     print(i,c)



#!/usr/bin/env python
from __future__ import print_function
from warnings import warn
from scipy import sparse
import _scs_direct

__version__ = _scs_direct.version()
__sizeof_int__ = _scs_direct.sizeof_int()
__sizeof_float__ = _scs_direct.sizeof_float()

_USE_INDIRECT_DEFAULT = False


def solve(probdata, cone, **kwargs):
  """Solves convex cone problems.

    @return dictionary with solution with keys:
         'x' - primal solution
         's' - primal slack solution
         'y' - dual solution
         'info' - information dictionary
  """
  if not probdata or not cone:
    raise TypeError('Missing data or cone information')

  if 'b' not in probdata or 'c' not in probdata:
    raise TypeError('Missing one or more of b, c from data dictionary')

  b = probdata['b']
  c = probdata['c']

  m = len(b)
  n = len(c)

  warm = {}
  if 'x' in probdata:
    warm['x'] = probdata['x']
  if 'y' in probdata:
    warm['y'] = probdata['y']
  if 's' in probdata:
    warm['s'] = probdata['s']

  if b is None or c is None:
    raise TypeError('Incomplete data specification')

  linsys_cbs = kwargs.get('linsys_cbs', None)
  if linsys_cbs:
    # Create an empty placeholder A matrix that is never used.
    A = sparse.csc_matrix((m,n))
  else:
    if 'A' not in probdata:
      raise TypeError('Missing A from data dictionary')
    A = probdata['A']

    if not sparse.issparse(A):
      raise TypeError('A is required to be a sparse matrix')
    if not sparse.isspmatrix_csc(A):
      warn('Converting A to a CSC (compressed sparse column) matrix; may take a '
           'while.')
      A = A.tocsc()

  if sparse.issparse(b):
    b = b.todense()

  if sparse.issparse(c):
    c = c.todense()

  Adata, Aindices, Acolptr = A.data, A.indices, A.indptr
  if kwargs.pop('gpu', False):  # False by default
    if not kwargs.pop('use_indirect', _USE_INDIRECT_DEFAULT):
      raise NotImplementedError(
          'GPU direct solver not yet available, pass `use_indirect=True`.')
    import _scs_gpu
    return _scs_gpu.csolve((m, n), Adata, Aindices, Acolptr, b, c, cone, warm,
                           **kwargs)

  if kwargs.pop('use_indirect', _USE_INDIRECT_DEFAULT):
    import _scs_indirect
    return _scs_indirect.csolve((m, n), Adata, Aindices, Acolptr, b, c, cone,
                                warm, **kwargs)

  if linsys_cbs:
    import _scs_python
    return _scs_python.csolve(
        (m, n), Adata, Aindices, Acolptr, b, c, cone,
        warm, **kwargs)

  return _scs_direct.csolve((m, n), Adata, Aindices, Acolptr, b, c, cone, warm,
                            **kwargs)



# nost test suite copied initially from ECOS project
from __future__ import print_function
import platform


def import_error(msg):
  print()
  print('## IMPORT ERROR:' + msg)
  print()


try:
  from nose.tools import assert_raises, assert_almost_equals
except ImportError:
  import_error('Please install nose to run tests.')
  raise

try:
  import scs
except ImportError:
  import_error('You must install the scs module before running tests.')
  raise

try:
  import numpy as np
except ImportError:
  import_error('Please install numpy.')
  raise

try:
  import scipy.sparse as sp
except ImportError:
  import_error('Please install scipy.')
  raise

# global data structures for problem
c = np.array([-1.])
b = np.array([1., -0.])
A = sp.csc_matrix([1., -1.]).T.tocsc()
data = {'A': A, 'b': b, 'c': c}
cone = {'q': [], 'l': 2}

FAIL = 'Failure'  # scs code for failure


def check_solution(solution, expected):
  assert_almost_equals(solution, expected, places=2)


def check_failure(sol):
  assert sol['info']['status'] == FAIL


def test_problems():
  sol = scs.solve(data, cone, use_indirect=False)
  yield check_solution, sol['x'][0], 1

  new_cone = {'q': [2], 'l': 0}
  sol = scs.solve(data, new_cone, use_indirect=False)
  yield check_solution, sol['x'][0], 0.5

  sol = scs.solve(data, cone, use_indirect=True)
  yield check_solution, sol['x'][0], 1

  sol = scs.solve(data, new_cone, use_indirect=True)
  yield check_solution, sol['x'][0], 0.5


if platform.python_version_tuple() < ('3', '0', '0'):

  def test_problems_with_longs():
    new_cone = {'q': [], 'l': long(2)}
    sol = scs.solve(data, new_cone, use_indirect=False)
    yield check_solution, sol['x'][0], 1
    sol = scs.solve(data, new_cone, use_indirect=True)
    yield check_solution, sol['x'][0], 1

    new_cone = {'q': [long(2)], 'l': 0}
    sol = scs.solve(data, new_cone, use_indirect=False)
    yield check_solution, sol['x'][0], 0.5
    sol = scs.solve(data, new_cone, use_indirect=True)
    yield check_solution, sol['x'][0], 0.5


def check_keyword(error_type, keyword, value):
  assert_raises(error_type, scs.solve, data, cone, **{keyword: value})


def test_failures():
  yield assert_raises, TypeError, scs.solve
  yield assert_raises, ValueError, scs.solve, data, {'q': [4], 'l': -2}
  # disable this until win64 types figured out
  # yield check_keyword, ValueError, 'max_iters', -1
  # python 2.6 and before just cast float to int
  if platform.python_version_tuple() >= ('2', '7', '0'):
    yield check_keyword, TypeError, 'max_iters', 1.1

  yield check_failure, scs.solve(data, {'q': [1], 'l': 0})
